/*
 * tests.cpp
 *
 *  Created on: Oct 14, 2008
 *  Revised Oct 2009
 *
 *  In this version, we define List, EmptyList, and NonEmptyList as templates which can be parameterized
 *  by the type of data to be held in the list.  Thus we can declare a variable as List<int>, List<float>, List<string>,
 *  even List<List<string>>.
 *  The syntax to do this is a pain: there are lots of things that you have to do in a certain way, not for any
 *  clear reason but because it won't compile otherwise.  In particular, standard procedure for template classes
 *  is to put the definitions of the member functions IN the include file, rather than having a separate .cpp file.
 *  This is because a template class on its own doesn't actually get compiled, so if it were in a separate .cpp file
 *  it wouldn't generate any object code anyway.
 *
 *  Once you've decided to put the definitions of member functions in the include file, you have a choice between
 *  putting them inside the curly-braces of the class definition, or putting them outside.  With ordinary classes,
 *  the main reason to put them outside was so they could be in a separate file.  That doesn't work very well for
 *  template classes, and the syntax is simpler if you put them inside the curly-braces as in Java, so that's what
 *  we'll usually do.
 *
 *      Author: bloch
 */

#include <iostream>
using std::cout;
using std::endl;
using std::string;

#include "List.h"
#include "EmptyList.h"
#include "NonEmptyList.h"



template<typename ElementType> NonEmptyList<ElementType> *force(List<ElementType> *blah)
{
	return dynamic_cast<NonEmptyList<ElementType> *>(blah);
}

typedef List<int> IntList;
typedef List<float> FloatList;
typedef List<string> StringList;
typedef EmptyList<int> EmptyIntList;
typedef EmptyList<float> EmptyFloatList;
typedef EmptyList<string> ESL;
typedef NonEmptyList<int> NonEmptyIntList;
typedef NonEmptyList<float> NonEmptyFloatList;
typedef NonEmptyList<string> NESL;

void testFloats(void);
void testInts(void);
void testStrings(void);

int main () {
	testFloats();
	testInts();
	testStrings();
	return 0;
}

void testFloats()
{
	FloatList *ex0 = new EmptyFloatList();
	FloatList *ex1 = new NonEmptyFloatList(3.2, ex0);
	FloatList *ex2 = new NonEmptyFloatList(-7.1, ex1);

	cout << "\nTesting floats:" << endl;
	cout << "ex0->isEmpty(): " << ex0->isEmpty() << " should be true" << endl;
	cout << "ex1->isEmpty(): " << ex1->isEmpty() << " should be false" << endl;
	cout << "ex2->isEmpty(): " << ex2->isEmpty() << " should be false" << endl;

	cout << "ex1->getFirst(): " << dynamic_cast<NonEmptyFloatList *>(ex1)->getFirst() << " should be 3.2 " << endl;

	// That's a pain.  Let's put that cast into a function:
	NonEmptyFloatList *ex1again = force(ex1);
	NonEmptyFloatList *ex2again = force(ex2);

	cout << "ex1again->getFirst(): " << ex1again->getFirst() << " should be 3.2" << endl;
	cout << "ex2again->getFirst(): " << ex2again->getFirst() << " should be -7.1" << endl;
	cout << "ex2again->getRest()->getFirst(): " <<
		force(ex2again->getRest())->getFirst() << " should be 3.2" << endl;
	cout << "ex0->length(): " << ex0->length() << " should be 0" << endl;
	cout << "ex1->length(): " << ex1->length() << " should be 1" << endl;
	cout << "ex2->length(): " << ex2->length() << " should be 2" << endl;

	cout << "ex0: " << (*ex0) << " should be []" << endl;
	cout << "ex1: " << (*ex1) << " should be [3.2]" << endl;
	cout << "ex2: " << (*ex2) << " should be [-7.1, 3.2]" << endl;

	FloatList *bigex =
		new NonEmptyFloatList (12.73,
				new NonEmptyFloatList (-2.851,
						new NonEmptyFloatList (-73.03,
								new NonEmptyFloatList (0,
										new NonEmptyFloatList (3.843,
												new EmptyFloatList())))));
	cout << "bigex->getRest()->getRest()->getFirst(): " <<
		force(force(force(bigex)->getRest())->getRest())->getFirst() <<
		" should be -73.03" << endl;
	cout << "bigex: " << *bigex << " should be [12.73, -2.851, -73.03, 0, 3.843]" << endl;
	cout << "bigex->length(): " << bigex->length() << " should be 5" << endl;

	/*
	cout << "ex1again->getRest()->getFirst(): " <<
		forceFloat(ex1again->getRest())->getFirst() <<
		" should throw an exception because ex1again->getRest() is empty" << endl;
	cout << "We shouldn't get here at all!" << endl;
	*/
}

void testInts()
{
	IntList *ex0 = new EmptyIntList();
	IntList *ex1 = new NonEmptyIntList(3, ex0);
	IntList *ex2 = new NonEmptyIntList(-7, ex1);

	cout << "\nTesting ints:" << endl;
	cout << "ex0->isEmpty(): " << ex0->isEmpty() << " should be true" << endl;
	cout << "ex1->isEmpty(): " << ex1->isEmpty() << " should be false" << endl;
	cout << "ex2->isEmpty(): " << ex2->isEmpty() << " should be false" << endl;

	cout << "ex1->getFirst(): " << dynamic_cast<NonEmptyIntList *>(ex1)->getFirst() << " should be 3 " << endl;

	// That's a pain.  Let's put that cast into a function:
	NonEmptyIntList *ex1again = force(ex1);
	NonEmptyIntList *ex2again = force(ex2);

	cout << "ex1again->getFirst(): " << ex1again->getFirst() << " should be 3" << endl;
	cout << "ex2again->getFirst(): " << ex2again->getFirst() << " should be -7" << endl;
	cout << "ex2again->getRest()->getFirst(): " <<
		force(ex2again->getRest())->getFirst() << " should be 3" << endl;
	cout << "ex0->length(): " << ex0->length() << " should be 0" << endl;
	cout << "ex1->length(): " << ex1->length() << " should be 1" << endl;
	cout << "ex2->length(): " << ex2->length() << " should be 2" << endl;

	cout << "ex0: " << (*ex0) << " should be []" << endl;
	cout << "ex1: " << (*ex1) << " should be [3]" << endl;
	cout << "ex2: " << (*ex2) << " should be [-7, 3]" << endl;

	IntList *bigex =
		new NonEmptyIntList (12,
				new NonEmptyIntList (-2,
						new NonEmptyIntList (-73,
								new NonEmptyIntList (0,
										new NonEmptyIntList (3,
												new EmptyIntList())))));
	cout << "bigex->getRest()->getRest()->getFirst(): " <<
		force(force(force(bigex)->getRest())->getRest())->getFirst() <<
		" should be -73" << endl;
	cout << "bigex: " << *bigex << " should be [12, -2, -73, 0, 3]" << endl;
	cout << "bigex->length(): " << bigex->length() << " should be 5" << endl;

/*
	cout << "ex1again->getRest()->getFirst(): " <<
		forceInt(ex1again->getRest())->getFirst() <<
		" should throw an exception because ex1again->getRest() is empty" << endl;
	cout << "We shouldn't get here at all!" << endl;
	*/
}


void testStrings()
{
	StringList *ex0 = new ESL();
	StringList *ex1 = new NESL("bye", ex0);
	StringList *ex2 = new NESL("hi", ex1);

	cout << "\nTesting strings:" << endl;
	cout << "ex0->isEmpty(): " << ex0->isEmpty() << " should be true" << endl;
	cout << "ex1->isEmpty(): " << ex1->isEmpty() << " should be false" << endl;
	cout << "ex2->isEmpty(): " << ex2->isEmpty() << " should be false" << endl;

	cout << "ex1->getFirst(): " << dynamic_cast<NESL *>(ex1)->getFirst() << " should be bye " << endl;

	// That's a pain.  Let's put that cast into a function:
	NESL *ex1again = force(ex1);
	NESL *ex2again = force(ex2);

	cout << "ex1again->getFirst(): " << ex1again->getFirst() << " should be bye" << endl;
	cout << "ex2again->getFirst(): " << ex2again->getFirst() << " should be hi" << endl;
	cout << "ex2again->getRest()->getFirst(): " <<
		force(ex2again->getRest())->getFirst() << " should be bye" << endl;
	cout << "ex0->length(): " << ex0->length() << " should be 0" << endl;
	cout << "ex1->length(): " << ex1->length() << " should be 1" << endl;
	cout << "ex2->length(): " << ex2->length() << " should be 2" << endl;

	cout << "ex0: " << (*ex0) << " should be []" << endl;
	cout << "ex1: " << (*ex1) << " should be [bye]" << endl;
	cout << "ex2: " << (*ex2) << " should be [hi, bye]" << endl;

	NESL *ex2copy = new NESL("hi", new NESL("bye", new ESL()));
	cout << "ex2copy: " << (*ex2copy) << " should be [hi, bye]" << endl;

	cout << "ex1 == ex2: " << (*ex1 == *ex2) << " should be false" << endl;
	cout << "ex2 == ex2: " << (*ex2 == *ex2) << " should be true" << endl;
	cout << "ex2 == new version of ex2: " << (*ex2 == *ex2copy) << " should be true" << endl;

	cout << "ex1 != ex2: " << (*ex1 != *ex2) << " should be true" << endl;
	cout << "ex2 != ex2: " << (*ex2 != *ex2) << " should be false" << endl;
	cout << "ex2 != new version of ex2: " << (*ex2 != *ex2copy) << " should be false" << endl;

	StringList *dwarfs =
		new NESL ("sleepy",
				new NESL ("sneezy",
						new NESL ("dopey",
								new NESL ("doc",
										new NESL ("happy",
												new NESL("bashful",
														new NESL("grumpy",
																new ESL())))))));
	cout << "dwarfs->getRest()->getRest()->getFirst(): " <<
		force(force(force(dwarfs)->getRest())->getRest())->getFirst() <<
		" should be dopey" << endl;
	cout << "dwarfs: " << *dwarfs << " should be [sleepy, sneezy, dopey, doc, happy, bashful, grumpy]" << endl;
	cout << "dwarfs->length(): " << dwarfs->length() << " should be 7" << endl;

/*
	cout << "ex1again->getRest()->getFirst(): " <<
		forceInt(ex1again->getRest())->getFirst() <<
		" should throw an exception because ex1again->getRest() is empty" << endl;
	cout << "We shouldn't get here at all!" << endl;
	*/
}
